use strict;
use vars qw(%RAD_REQUEST %RAD_REPLY %RAD_CHECK);

use Authen::Radius;
use XML::Simple;

use Data::Dumper;

use constant    RLM_MODULE_REJECT=>    0; #  /* immediately reject the request */
use constant    RLM_MODULE_FAIL=>      1; #  /* module failed, don't reply */
use constant    RLM_MODULE_OK=>        2; #  /* the module is OK, continue */
use constant    RLM_MODULE_HANDLED=>   3; #  /* the module handled the request, so stop. */
use constant    RLM_MODULE_INVALID=>   4; #  /* the module considers the request invalid. */
use constant    RLM_MODULE_USERLOCK=>  5; #  /* reject the request (user is locked out) */
use constant    RLM_MODULE_NOTFOUND=>  6; #  /* user not found */
use constant    RLM_MODULE_NOOP=>      7; #  /* module succeeded without doing anything */
use constant    RLM_MODULE_UPDATED=>   8; #  /* OK (pairs modified) */
use constant    RLM_MODULE_NUMCODES=>  9; #  /* How many return codes there are */

# get Config:
my $conf = XMLin("/etc/freeradius/perl/auth-token-remote.xml");
$conf->{radius} = new Authen::Radius(
    Host => $conf->{"radius-server"}->{"host"},
    Secret => $conf->{"radius-server"}->{"secret"} );

print STDERR "auth-token-remote: init done\n";
print STDERR Dumper($conf) . "\n"; # DEBUG

sub authenticate {
  print "rlm_perl AUTHENTICATE\n";

  my $tokenId = $RAD_CHECK{"Acc-Ace-Token"};

#  print "auth-token-remote: CHECK: " . Dumper(\%RAD_CHECK) . "\n"; # DEBUG
#  print "auth-token-remote: REQUEST: " . Dumper(\%RAD_REQUEST) . "\n"; # DEBUG
  
  if (!defined($tokenId)) {
      print "auth-token-remote: no Acc-Ace-Token defined\n";
      return RLM_MODULE_OK;
  }

  my ($userPassword,$tokenVal) = ($RAD_REQUEST{"User-Password"} =~ /^(.*)(.{6})$/);

  my $remotePass;
  if ($RAD_CHECK{"Cleartext-Password"} eq "REMOTE") {
      $remotePass = $RAD_REQUEST{"User-Password"};

      # Make local PAP auth just pass through:
      $RAD_REQUEST{"User-Password"} = $RAD_CHECK{"Cleartext-Password"};
  } else {
      $remotePass = $tokenVal;
      $RAD_REQUEST{"User-Password"} = $userPassword;
  }

  if (!defined($userPassword)) {
      $RAD_REQUEST{'Reply-Message'} = "auth-token-remote: missing pw";
      print STDERR "auth-token-remote: missing pw\n";
      return RLM_MODULE_REJECT;
  }

  if (!defined($remotePass) || $remotePass eq "") {
      $RAD_REQUEST{'Reply-Message'} = "auth-token-remote: missing 6-digit token";
      print STDERR "auth-token-remote: missing token\n";
      return RLM_MODULE_REJECT;
  }

  print "remoteauth $tokenId, $remotePass\n";
  my $auth = $conf->{radius}->check_pwd($tokenId, $remotePass);

#  print "REMOTE REPLY: " . Dumper($r->get_attributes()) . "\n"; # DEBUG

  if (!$auth) {
      $RAD_REQUEST{'Reply-Message'} = "auth-token-remote: auth failed";
      print STDERR "auth-token-remote: remote auth failed\n";
      return RLM_MODULE_REJECT;
  }

  print STDERR "auth-token-remote: auth ok\n";

  return RLM_MODULE_OK;
}


1;



