#
# Wed Mar  2 16:05:16 CET 2016 -Rodney
# - Ueberarbeitung.
#
#
use strict;
use warnings;

use vars qw(%RAD_REQUEST %RAD_REPLY %RAD_CHECK);

use Authen::OATH;
#use LWP::UserAgent;
use Data::Dumper;

use constant    RLM_MODULE_REJECT=>    0;#  /* immediately reject the request */
use constant    RLM_MODULE_FAIL=>      1;#  /* module failed, don't reply */
use constant    RLM_MODULE_OK=>        2;#  /* the module is OK, continue */
use constant    RLM_MODULE_HANDLED=>   3;#  /* the module handled the request, so stop. */
use constant    RLM_MODULE_INVALID=>   4;#  /* the module considers the request invalid. */
use constant    RLM_MODULE_USERLOCK=>  5;#  /* reject the request (user is locked out) */
use constant    RLM_MODULE_NOTFOUND=>  6;#  /* user not found */
use constant    RLM_MODULE_NOOP=>      7;#  /* module succeeded without doing anything */
use constant    RLM_MODULE_UPDATED=>   8;#  /* OK (pairs modified) */
use constant    RLM_MODULE_NUMCODES=>  9;#  /* How many return codes there are */

my $oath = Authen::OATH->new(
    digits => 6,
    timestep => 60,
    );

print STDERR "auth-token: initdone\n";

my @tokenMem;

sub authenticate {
    logit("authenticate()");

    print Dumper(\%RAD_REQUEST) . "\n";

    # FIXME: Should allowed usernames/passwords be limited to a certain regex?
    my ($req_username) = ($RAD_REQUEST{'User-Name'} =~ /^(.*)$/);
    my $req_userpassword = $RAD_REQUEST{'User-Password'};
    
    if (!defined($req_username)) {
	logit("missing 'User-Name' in request");
	return(RLM_MODULE_REJECT);
    }
    
    if (!defined($req_userpassword)) {
	logit("missing 'User-Password' in request");
	return(RLM_MODULE_REJECT);
    }


    # ------------------------------------------------------------------------------------
    # Case: $RAD_CHECK{'Acc-Ace-Token'} exists for this user
    # - we expect the user to supply a token as the last 6 digits of the password
    # - we set $tokenVal accordingly and strip it off of $userPassword
    #
    my $req_tokenval = undef;

    my $check_tokenkey = $RAD_CHECK{'Acc-Ace-Token'};
    if (defined($check_tokenkey)) {
	if (my ($a, $b) = ($req_userpassword =~ /^(.*)(\d{6})$/)) {
	    $req_userpassword = $a // "";
	    $req_tokenval = $b;
	}
    }


    # Case: Check only password, without token. ------------------------------------------
    if (defined(my $check_cleartextpassword = $RAD_CHECK{"Cleartext-Password"})) {
	if ($check_cleartextpassword ne $req_userpassword) {
	    logit("'Cleartext-Password' is set, but FAILED check for supplied PW.");
	    return(RLM_MODULE_REJECT);
	}
	logit("pw check passed.");
    }


    # Send password via SMS text message service ---------------------------------------
    #
    # We send a token via SMS, if all of the following are true:
    #
    # - Acc-Ace-Token is configured, i. e. we expect an OTP token, and
    # - No OTP token was supplied, and
    # - The password check was successful, and
    # - A "Callback-Number" is configured.
    #
    # "Callback-Number" is expected to be a URL, e.g.:
    #
    #  https://monitor.lf.net/sms.php?nr=01727114238&t=$pw
    #
    if (($check_tokenkey // "") ne "" &&
	($req_tokenval // "") eq "" &&
	defined(my $reply_callbacknumber = $RAD_REPLY{"Callback-Number"})) {
	
	my $url = $reply_callbacknumber;
	my $thisKey = $oath->totp($check_tokenkey, time());
	$url =~ s/\$pw/$thisKey/;
	logit("sending password via sms to: $url");
	system("curl -k -3 '$url'"); # FIXME, clumsy.

	return(RLM_MODULE_REJECT);
    }

    # The user should not see internal data.
    undef($RAD_REPLY{"Callback-Number"});


#    if (!defined($tokenVal)) {
#	logit("missing user-password in request");
#	return(RLM_MODULE_REJECT);
#    }

#    # FIXME: Description for this step!
#    if ($RAD_CHECK{"Cleartext-Password"} eq "REMOTE") {
#	$userPassword = "REMOTE";
#	$tokenVal = $RAD_REQUEST{'User-Password'};
#    }

#    # We expect a Acc-Ace-Token key to be present in our configuration.
#    # FIXME: we should also allow password-only users, and do the pw management
#    # for them.
#    my $tokenKey = $RAD_CHECK{'Acc-Ace-Token'};
#    if (!defined($tokenKey)) {
#	logit("no tokenKey found");
#	return(RLM_MODULE_REJECT);
#    }

    my $t = time();
#    $t += 80; # DEBUG


    if ($check_tokenkey) {
	my $authSuccess = 0;

	if (($req_tokenval // "") eq "") {
	    logit("FAILED because check 'Acc-Ace-Token' is set, but no token received.");
	    return(RLM_MODULE_REJECT);
	}

	for (my $delta = -120; $delta < 120; $delta += 10) {
	    my $otp = $oath->totp($check_tokenkey, $t + $delta);
	    if ($otp eq $req_tokenval) {
		$authSuccess = 1;
		logit("token verification good for $req_username, delta=$delta");
		last;
	    }
	    
	    # Google Authenticator on android does not honour period=n.
	    # Needs $t * 2, since calculation is done based on unixtime() mod 30.
	    $otp = $oath->totp($check_tokenkey, $t * 2 + $delta);
	    if ($otp eq $req_tokenval) {
		$authSuccess = 1;
		logit("token verification (1/2) good for $req_username, delta=$delta");
		last;
	    }
	}
	
	#    $authSuccess = 1; # DEBUG
	
	if (!$authSuccess) {
	    logit("token verification FAILED ($req_username, OTP req_tokenval=$req_tokenval)");
	    return(RLM_MODULE_REJECT);
	}
	

	# Make sure OTP passwords can only be used once --------------------------------
       	# Expire keys from keystore after n seconds:
	@tokenMem = grep { $_->{"time"} > $t - 300 } @tokenMem;
	
	# For the tuple ($tokenKey, $tokenVal) passwords shall be allowed only once.
	my $tokenMemKey = $req_username . "." . $req_tokenval;
	
	# Check if this key was already used:
	my @used = grep { $_->{"key"} eq $tokenMemKey} @tokenMem;
	
	if ($#used > -1) {
	    logit("FAILED: token already used");
	    return(RLM_MODULE_REJECT);
	}
	
	# Remember that this key was used.
	push(@tokenMem, { "key" => $tokenMemKey, "time" => $t });
	
	logit("token auth successful (t=$t, otp=$req_tokenval)");
    }
    
    return RLM_MODULE_OK;
}



sub logit {
    my ($s) = @_;
    print STDERR "auth-token: $s\n";
}

1;
