use strict;
use warnings;

use vars qw(%RAD_REQUEST %RAD_REPLY %RAD_CHECK);

use Authen::OATH;
#use LWP::UserAgent;
use Data::Dumper;

use constant    RLM_MODULE_REJECT=>    0;#  /* immediately reject the request */
use constant    RLM_MODULE_FAIL=>      1;#  /* module failed, don't reply */
use constant    RLM_MODULE_OK=>        2;#  /* the module is OK, continue */
use constant    RLM_MODULE_HANDLED=>   3;#  /* the module handled the request, so stop. */
use constant    RLM_MODULE_INVALID=>   4;#  /* the module considers the request invalid. */
use constant    RLM_MODULE_USERLOCK=>  5;#  /* reject the request (user is locked out) */
use constant    RLM_MODULE_NOTFOUND=>  6;#  /* user not found */
use constant    RLM_MODULE_NOOP=>      7;#  /* module succeeded without doing anything */
use constant    RLM_MODULE_UPDATED=>   8;#  /* OK (pairs modified) */
use constant    RLM_MODULE_NUMCODES=>  9;#  /* How many return codes there are */

die("This is obsoleted by auth-token-v2.\n");

my $oath = Authen::OATH->new(
    digits => 6,
    timestep => 60,
    );

print STDERR "auth-token: initdone\n";

my @tokenMem;

sub authenticate {
    logit("authenticate()");

    my $tokenId = $RAD_REQUEST{'User-Name'} || "";
    my ($userPassword, $tokenVal) = ($RAD_REQUEST{'User-Password'} =~ /^(.*)(.{6})$/);

    print Dumper(\%RAD_REQUEST) . "\n";
    print "userpassword '$userPassword'\n";

    if (!defined($tokenVal)) {
	logit("missing user-password in request");
	return(RLM_MODULE_REJECT);
    }

    # FIXME: Description for this step!
    if ($RAD_CHECK{"Cleartext-Password"} eq "REMOTE") {
	$userPassword = "REMOTE";
	$tokenVal = $RAD_REQUEST{'User-Password'};
    }

    # We expect a Acc-Ace-Token key to be present in our configuration.
    # FIXME: we should also allow password-only users, and do the pw management
    # for them.
    my $tokenKey = $RAD_CHECK{'Acc-Ace-Token'};
    if (!defined($tokenKey)) {
	logit("no tokenKey found");
	return(RLM_MODULE_REJECT);
    }

    my $t = time();
#    $t += 80; # DEBUG

    # Send password via SMS text message service ---------------------------------------
    #
    # If a "Callback-Number" is given, and "User-Password" == SENDSMS,
    # we send the current TOTP password to the user.
    #
    # "Callback-Number" is expected to be a URL, e.g.:
    #
    #  https://monitor.lf.net/sms.php?nr=01727114238&t=$pw
    #
    print "repl: $RAD_REPLY{\"Callback-Number\"} upw: $RAD_REQUEST{'User-Password'}\n";
    if ($RAD_REPLY{"Callback-Number"} ne "" && 
	$RAD_REQUEST{"User-Password"} eq "SENDSMS") {

	my $url = $RAD_REPLY{"Callback-Number"};
	my $thisKey = $oath->totp($tokenKey, $t);
	$url =~ s/\$pw/$thisKey/;
	logit("sending password via sms to: $url");
	system("curl -3 '$url'"); # FIXME, clumsy.
	return(RLM_MODULE_REJECT);
    }

    $RAD_REQUEST{'User-Password'} = $userPassword;

    my $authSuccess = 0;
    for (my $delta = -120; $delta < 120; $delta += 10) {
	my $otp = $oath->totp($tokenKey, $t + $delta);
	if ($otp eq $tokenVal) {
	    $authSuccess = 1;
	    logit("auth success for tokenId=$tokenId delta=$delta");
	    last;
	}

	# Google Authenticator on android does not honour period=n.
	# Needs $t * 2, since calculation is done based on unixtime() mod 30.
	$otp = $oath->totp($tokenKey, $t * 2 + $delta);
	if ($otp eq $tokenVal) {
	    $authSuccess = 1;
	    logit("auth 1/2 success for tokenId=$tokenId delta=$delta");
	    last;
	}
    }

#    $authSuccess = 1; # DEBUG

    if (!$authSuccess) {
	logit("auth failed ($tokenId, $tokenVal)");
	return(RLM_MODULE_REJECT);
    }


    # Make sure OTP passwords can only be used once --------------------------------

    # Expire keys from keystore after n seconds:
    @tokenMem = grep { $_->{"time"} > $t - 300 } @tokenMem;

    # For the tuple ($tokenKey, $tokenVal) passwords shall be allowed only once.
    my $tokenMemKey = $tokenKey . "." . $tokenVal;

    # Check if this key was already used:
    my @used = grep { $_->{"key"} eq $tokenMemKey} @tokenMem;

    if ($#used > -1) {
	logit("token already used");
	return(RLM_MODULE_REJECT);
    }

    # Remember that this key was used.
    push(@tokenMem, { "key" => $tokenMemKey, "time" => $t });

    logit("auth successful (t=$t, otp=$tokenVal)");

    return RLM_MODULE_OK;
}

sub logit {
    my ($s) = @_;
    print STDERR "auth-token: $s\n";
}

1;
